package org.zeroturnaround.javarebel.integration.resin;

import org.zeroturnaround.bundled.javassist.CannotCompileException;
import org.zeroturnaround.bundled.javassist.ClassPool;
import org.zeroturnaround.bundled.javassist.CtClass;
import org.zeroturnaround.bundled.javassist.CtMethod;
import org.zeroturnaround.bundled.javassist.CtNewMethod;
import org.zeroturnaround.bundled.javassist.NotFoundException;
import org.zeroturnaround.bundled.javassist.expr.ExprEditor;
import org.zeroturnaround.bundled.javassist.expr.MethodCall;
import org.zeroturnaround.javarebel.integration.support.JavassistClassBytecodeProcessor;

/**
 * Transforms <code>com.caucho.loader.DynamicClassLoader</code>.
 * 
 * <p>Tested on Resin 3.0.27, 3.1.3, 3.2.1 and 4.0.0.</p>
 * 
 * @author Rein Raudjärv
 */
public class ResinDynamicClassLoaderCBP extends JavassistClassBytecodeProcessor {
  
  public void process(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    cp.importPackage("java.io");
    cp.importPackage("java.net");
    cp.importPackage("java.util");
    cp.importPackage("com.caucho.loader");
    cp.importPackage("org.zeroturnaround.javarebel");
    cp.importPackage("org.zeroturnaround.javarebel.integration.generic");
    
    /*
     * Register class loader (except a SystemClassLoader)
     */
    ctClass.getDeclaredMethod("init").insertAfter(
          "{" +
          "  if (!(this instanceof SystemClassLoader)) {" +
          "    IntegrationFactory.getInstance().registerClassLoader($0, new FindResourceClassResourceSource($0));" +
          "  }" +
          "}");
    
    /*
     * Notice JavaRebel if class loader has new URLS attached.
     * 
     * Required for Resin 4.0.x. Prior that not sure.
     */
    try {
      ctClass.getDeclaredMethod("addURL", cp.get(new String[] { "int", "java.net.URL" })).insertAfter(
          "{ " + 
          "  IntegrationFactory.getInstance().reinitializeClassLoader($0);" +
          "}");
    } catch (NotFoundException e) { /* skip */ }    
    
    processClassLoading(ctClass);
    
    processResources(cp, ctClass);
  }

  private void processClassLoading(CtClass ctClass) throws CannotCompileException, NotFoundException {
    /*
     * Override class loading.
     */
    String methodName = "findClass";
    try {
      ctClass.getDeclaredMethod("findClassImpl");
      methodName = "findClassImpl";
    } catch (NotFoundException e) {}

    ctClass.getDeclaredMethod(methodName).insertBefore(
        "{ synchronized ($0) {" + 
        "    Class result =" + 
        "      findLoadedClass($1);" + 
        "    if (result != null)" + 
        "      return result;" + 
        "    result = " + 
        "      IntegrationFactory.getInstance().findReloadableClass($0, $1);" + 
        "    if (result != null)" + 
        "      return result;" + 
    "}}");
  }

  private void processResources(ClassPool cp, CtClass ctClass) throws CannotCompileException, NotFoundException {
    /*
     * Add an empty method - getImportResource() - if missing. 
     */
    try {
      // getImportResource was added in Resin 3.2 for resources from OSGI
      ctClass.getDeclaredMethod("getImportResource");
    } catch (NotFoundException e) {
      ctClass.addMethod(CtNewMethod.make(
          "protected URL getImportResource(String name) {" +
          "  return null;" +
          "}", ctClass));
    }
    
    ctClass.addMethod(CtNewMethod.make(
        "private URL __getLocalResource(String name) {" +

        "  Integration integration = IntegrationFactory.getInstance();" +
        "  if (integration.isResourceReplaced($0, name))" +
        "    return integration.findResource($0, name);" +
        
        "  URL url = getImportResource(name);" +
        "  if (url != null)" +
        "    return url;" +

        "  ArrayList loaders = getLoaders();" +
        "  for (int i = 0; loaders != null && i < loaders.size(); i++) {" +
        "    Loader loader = (Loader) loaders.get(i);" +
        "    url = loader.getResource(name);" +
        "    if (url != null)" +
        "      return url;" +
        "  }" +

        "  return null;" +

        "}", ctClass));
    
    ctClass.addMethod(CtNewMethod.make(
        "private InputStream __getLocalResourceAsStream(String name) {" +

        "  Integration integration = IntegrationFactory.getInstance();" +
        "  if (integration.isResourceReplaced($0, name)) {" +
        "    URL url = integration.findResource($0, name);" +

        "    if (url == null)" +
        "      return null;" +
        
        "    try {" +
        "      return url.openStream();" +
        "    } catch (IOException e) {" +
        "      return null;" +
        "    }" +
        "  }" +
        
        "  ArrayList loaders = getLoaders();" +
        "  for (int i = 0; loaders != null && i < loaders.size(); i++) {" +
        "    Loader loader = (Loader) loaders.get(i);" +
        "    InputStream is = loader.getResourceAsStream(name);" +
        "    if (is != null)" +
        "      return is;" +
        "  }" +

        "  return null;" +

        "}", ctClass));    

    try {
      CtMethod m = ctClass.getDeclaredMethod("getResource");
      m.setBody("{" +
          "  if ($1.startsWith(\"/\"))" +
          "    $1 = $1.substring(1);" +

          "  if ($1.endsWith(\"/\"))" +
          "    $1 = $1.substring(0, $1.length() - 1);" +

          "  boolean isNormalJdkOrder = isNormalJdkOrder($1);" +
          "  URL url;" +
          
          "  if (isNormalJdkOrder) {" +
          "    url = getParentResource($1);" +
          "    if (url != null)" +
          "      return url;" +
          "  }" +
          
          "  url = __getLocalResource($1);" +
          "  if (url != null)" +
          "    return url;" +

          "  if (! isNormalJdkOrder) {" +
          "    url = getParentResource($1);" +
          "    if (url != null)" +
          "      return url;" +
          "  }" +

          "  return null;" +
          "}");
    }
    catch (NotFoundException e) {
      // skip
    }
    
    try {
      CtMethod m = ctClass.getDeclaredMethod("getResourceAsStream");
      m.setBody("{" +
          "  if ($1.startsWith(\"/\"))" +
          "    $1 = $1.substring(1);" +

          "  if ($1.endsWith(\"/\"))" +
          "    $1 = $1.substring(0, $1.length() - 1);" +
          
          "  boolean isNormalJdkOrder = isNormalJdkOrder($1);" +
          "  InputStream is;" +
          
          "  if (isNormalJdkOrder) {" +
          "    is = getParentResourceAsStream($1);" +
          "    if (is != null)" +
          "      return is;" +
          "  }" +
          
          "  is = __getLocalResourceAsStream($1);" +
          "  if (is != null)" +
          "    return is;" +
          
          "  if (! isNormalJdkOrder) {" +
          "    is = getParentResourceAsStream($1);" +
          "    if (is != null)" +
          "      return is;" +
          "  }" +
          
          "  return null;" +
      "}");
    }
    catch (NotFoundException e) {
      // skip
    }

    try {
      CtMethod m = ctClass.getDeclaredMethod("findResources", cp.get(new String[] {String.class.getName()}));
      m.insertBefore(
          "{" + 
          "  String name = $1;" +
          "  if (name.startsWith(\"/\"))" +
          "    name = name.substring(1);" +

          "  Integration integration = IntegrationFactory.getInstance();" +
          "  if (integration.isResourceReplaced($0, name))" +
          "    return integration.findResources($0, name);" + 
          "}");
    }
    catch (NotFoundException e) {
      // skip
    }
    
    /*
     * Disable cache usage in method getParentResource()
     */
    ctClass.getDeclaredMethod("getParentResource").instrument(new ExprEditor() {
      public void edit(MethodCall m) throws CannotCompileException {
        if (m.getClassName().equals("com.caucho.util.TimedCache")
            && m.getMethodName().equals("put")) {
          m.replace("{ $_ = null; }");
        }
      }
    });
  }
}
